# -*- coding: utf-8 -*-

import os
import random
import re
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from logging import basicConfig, getLogger
logger = getLogger(__name__)

from .base import WordleAI
from .utils import show_word_evaluations, default_wordle_vocab, _timereport, wordle_judge, decode_judgement, _read_vocabfile
from .sqlite import WordleAISQLite
from .approx import WordleAIApprox
from . import __version__

def interactive(ai: WordleAI, num_suggest: int=10, default_criterion: str="mean_entropy"):
    print("")
    print("Hi, this is %s." % ai.name)
    print("")
    #ai.set_candidates()  # initialize all candidates
    ai.clear_info()
    words_set = set(ai.words)

    def _receive_input(words_set: set):
        while True:
            message = [
                "",
                "Type:",
                "  '[s]uggest <criterion>'     to let AI suggest a word (<criterion> is optional)",
                "  '[u]pdate <word> <result>'  to provide new information",
                "  '[e]xit'                    to finish the session",
                "", 
                "where",
                "  <criterion>  is either 'max_n', 'mean_n', or 'mean_entropy'",
                "  <result>     is a string of 0 (no match), 1 (partial match), and 2 (exact match)",
                "",
                "> "
            ]
            ans = input("\n".join(message))
            ans = re.sub(r"\s+", " ", ans.strip())
            ans = ans.split(" ")
            if len(ans) <= 0: continue
            if len(ans[0]) <= 0: continue

            if ans[0][0] == "s":
                if len(ans) > 1:
                    criterion = ans[1]
                    if criterion not in ("max_n", "mean_n", "mean_entropy"):
                        print("Invalid <criterion> ('%s' is given)" % criterion)
                        continue
                    return ["s", criterion]
                else:
                    return ["s"]
            elif ans[0][0] == "u":
                if len(ans) < 3:
                    continue
                word, result = ans[1], ans[2]
                if not all(r in "012" for r in result):
                    print("'%s' is invalid result expression" % result)
                    continue
                if word not in words_set:
                    print("'%s' is not in the vocab" % word)
                    continue
                if len(word) < len(result):
                    print("Word and result length mismatch")
                    continue
                return ["u", word, result]
            elif ans[0][0] == "e":
                return ["e"]
            
    while True:
        maxn = 10  # max number of candidates to show
        cur_candidates = ai.candidates
        n_remain = len(cur_candidates)
        remain = cur_candidates[:maxn]
        if n_remain > maxn:
            remain.append("...")
        if n_remain > 1:
            print("%d remaining candidates: %s" % (n_remain, remain))
        elif n_remain==1:
            print("'%s' should be the answer!" % remain[0])
            break
        else:
            print("There is no candidate words consistent with the information...")
            break

        ans = _receive_input(words_set)
        if ans[0] == "s":
            criterion = default_criterion if len(ans) < 2 else ans[1]
            with _timereport("AI evaluation"):
                res = ai.evaluate(top_k=num_suggest, criterion=criterion)
            print("* Top %d candidates ordered by %s" % (len(res), criterion))
            show_word_evaluations(res)
        elif ans[0] == "u":
            ai.update(ans[1], ans[2])
        elif ans[0] == "e":
            break
    print("Thank you!")


def play(words: list or dict, vocabname: str="No name"):
    if isinstance(words, list):
        words = {w:1 for w in words}  # assign equal weight

    tmp = list(words)[:5]
    if len(words) > 5:
        tmp.append("...")
    print("")
    print("Enjoy wordle game (vocabname: '%s', containing %d words, e.g. %s)" % (vocabname, len(words), tmp))
    print("")
    print("Type your guess, or 'give up' to finish the game")

    # pick an answer randomly
    vals = []
    weights = []
    for w, p in words.items():
        if p > 0:
            vals.append(w)
            weights.append(p)
    assert len(vals) > 0, "There is no word with positive weight"
    answer_word = random.choices(vals, weights, k=1)[0]
    wordlen = len(answer_word)
        
    # define a set version of words for quick check for existence
    input_words_set = set(words)
    def _get_word():
        while True:
            x = input("> ").strip()
            if x in input_words_set or x == "give up":
                return x
            print("Invalid word: '%s'" % x)
                
    round_ = 0
    info = []
    while True:
        round_ += 1
        print("* Round %d *" % round_)
        input_word = _get_word()
        if input_word == "give up":
            print("You lose. Answer: '%s'." % answer_word)
            return False
        res = wordle_judge(input_word, answer_word)
        res = str(decode_judgement(res)).zfill(wordlen)
        info.append("  %s  %s" % (input_word, res))
        print("\n".join(info))
        if input_word == answer_word:
            print("Good job! You win! Answer: '%s'" % answer_word)
            return True

def challenge(ai: WordleAI, max_round: int=20, visible: bool=False,
              alternate: bool=False, ai_first: bool=False, continue_after_result: bool=False):
    #ai.set_candidates()
    ai.clear_info()
    n_ans = len(ai.candidates)
    n_words = len(ai.words)

    tmp = ai.words[:5]
    if n_words > 5:
        tmp.append("...")
    print("")
    print("Wordle game against %s, AI strength: %s, vocabname: %s" % (ai.name, ai.strength, ai.vocabname))
    print("%d words, e.g. %s" % (n_words, tmp))
    print("")
    print("Type your guess, or 'give up' to finish the game")
    print("")

    # pick an answer randomly
    answer_word = ai.choose_answer_word()
    wordlen = len(answer_word)

    # define a set version of words for quick check for existence
    words_set = set(ai.words)
    def _get_word():
        while True:
            x = input("Your turn > ").strip()
            if x in words_set or x == "give up":
                return x
            print("Invalid word: '%s'" % x)

    user_history, ai_history = [], []
    user_done, ai_done = False, False  # flag of finding answer
    giveup = False  # flag user gave up
    ai_turn = ai_first
    winner = "unfinished"  # {unfinished, draw, user, ai}
    def _show_history():
        _user_history = user_history.copy()
        _ai_history = ai_history.copy()
        #print(_user_history)
        #print(_ai_history)
        # fix the legngth difference, if any
        if not visible:
            _ai_history = [(("*" if w in words_set else " ") * wordlen, r) for w, r in _ai_history]
        # pad empty rows
        while len(_ai_history) < len(_user_history):
            _ai_history.append((" " * wordlen, " " * wordlen))
        while len(_user_history) < len(_ai_history):
            _user_history.append((" " * wordlen, " " * wordlen))
        out = []
        for u, a in zip(_user_history, _ai_history):
            out.append("  %s  %s  |  %s  %s" % tuple(u + a))
        print("\n".join(out))

    for round_ in range(max_round):
        print("* Round %d *" % (round_ + 1))
        # clear prev round result
        tmp = " " * wordlen
        ai_word, ai_res, user_word, user_res = tmp, tmp, tmp, tmp
        for _ in range(2):  # loop over two users
            skipped = False  # flag no decision has been made by this user
            if ai_turn:
                if ai_done:
                    skipped = True
                else:
                    with _timereport("AI thinking"):
                        ai_word = ai.pick_word()
                    ai_res = wordle_judge(ai_word, answer_word)
                    ai_res = str(decode_judgement(ai_res)).zfill(wordlen)
                    ai.update(ai_word, ai_res)
                    ai_history.append((ai_word, ai_res))
                    if ai_word == answer_word:
                        ai_done = True
                ai_turn = False
            else:
                if (user_done or giveup):
                    skipped = True
                else:
                    user_word = _get_word()
                    if user_word == "give up":
                        giveup = True
                    else:
                        user_res = wordle_judge(user_word, answer_word)
                        user_res = str(decode_judgement(user_res)).zfill(wordlen)
                        user_history.append((user_word, user_res))
                        if user_word == answer_word:
                            user_done = True
                        if visible and alternate and (user_word in words_set):
                            # ai learns the user output
                            ai.update(user_word, user_res)
                ai_turn = True

            if alternate and (not skipped):
                _show_history() # show the history at the end of one move
                # check the winner
                if winner == "unfinished":
                    if ai_done:
                        winner = "ai"
                    elif user_done:
                        winner = "user"
                if winner != "unfinished":
                    if (not continue_after_result) or visible:
                        # visible --> answer is already seen
                        # not continue_after_result --> no reason to continue
                        break  # go to the round end

        if not alternate:
            _show_history()  # show the history at the end of each found
            if visible and (user_word in words_set): 
                # ai learns user's output at the end of round
                ai.update(user_word, user_res)
            # check the winner
            if winner == "unfinished":
                if ai_done and user_done:
                    winner = "draw"
                elif ai_done:
                    winner = "ai"
                elif user_done:
                    winner = "user"

        if continue_after_result and (not visible):
            if user_done and ai_done:
                break
        else:
            if winner != "unfinished":
                break
    print("===============================")
    if winner == "user":
        print("Great job! You win!")
    elif winner == "ai":
        print("You lose...")
    else:
        print("Good job. It's draw.")
    print("Answer: '%s'" % answer_word)
    visible = True
    _show_history()
    print("===============================")
            

def main():
    parser = ArgumentParser(description="Wordle AI with SQL backend", formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-b", "--backend", type=str, default="approx", choices=["sqlite", "approx", "bq", "random"], help="AI type")
    parser.add_argument("--vocabname", default=None, type=str, help="Name of vocabulary")
    parser.add_argument("--vocabfile", type=str, help="Text file containing words. If not supplied, default wordle vocab is used")
    parser.add_argument("--resetup", action="store_true", help="Setup the vocabulary if already exists")
    parser.add_argument("--sqlitefile", type=str, 
                        help=("SQLite database file. If not supplied, we first search env variable 'WORDLEAISQL_DBFILE'. "
                              "If the env variable is not defined, then ./wordleai.db is used"))
    parser.add_argument("--inmemory", action="store_true", help="Use in-memory database. Only applicable with `-b approx`")
    parser.add_argument("--word_pair_limit", type=int, default=500000,
                        help="Maximum number of (input word, answer word) pairs computed for approximate evaluation")
    parser.add_argument("--candidate_samplesize", type=int, default=500,
                        help="Sample size of answer word for approximate evaluation")
    parser.add_argument("--bq_credential", type=str, help="Credential json file for a GCP service client")
    parser.add_argument("--bq_project", type=str, help="GCP project id (if not supplied, inferred from the credential default)")
    parser.add_argument("--bq_location", type=str, default="US", help="GCP location")
    parser.add_argument("--partition_size", type=int, default=200, help="Partition size of judges table")

    parser.add_argument("--suggest_criterion", type=str, default="mean_entropy", choices=["max_n", "mean_n", "mean_entropy"],
                        help="Criterion for an AI to sort the word suggestions")
    parser.add_argument("--num_suggest", type=int, default=20, help="Number of suggestion to print")

    parser.add_argument("--play", action="store_true", help="Play your own game without AI")
    parser.add_argument("--answer_difficulty", type=int, default=3, choices=[1,2,3,4,5], help="Answer word difficulty from 1 (basic) to 5 (unlimited)")
    parser.add_argument("--challenge", action="store_true", help="Challenge AI")
    parser.add_argument("--max_round", type=int, default=20, help="Maximum rounds in challenge mode")
    parser.add_argument("--visible", action="store_true", help="Opponent words are visible in challenge mode")
    parser.add_argument("--alternate", action="store_true", help="Decisions are made in turn in challenge mode")
    parser.add_argument("--ai_first", action="store_true", help="AI makes the first decision in challenge mode")
    parser.add_argument("--continue_after_result", action="store_true", help="Continue the game after result is determined in challenge mode")
    parser.add_argument("--ai_strength", type=float, default=6, help="Strength of AI in [0, 10] in challenge mode")
    parser.add_argument("--decision_metric", type=str, default="mean_entropy", choices=["max_n", "mean_n", "mean_entropy"],
                        help="Criterion for an AI to use in challenge mode")
    parser.add_argument("--candidate_weight", type=float, default=0.3, help="Weight applied to the answer candidate words in challenge mode")

    parser.add_argument("--no_cpp", action="store_true", help="Not to use C++ script even if available")
    parser.add_argument("--cpp_recompile", action="store_true", help="Compile the C++ script again even if the source script is not updated")
    parser.add_argument("--cpp_compiler", type=str, help="Command name of the C++ compiler")

    parser.add_argument("--debug", action="store_true", help="Show debug messages")
    parser.add_argument("--version", action="store_true", help="Show the program version")

    args = parser.parse_args()
    #print(args)
    if args.version:
        print("wordleaisql v%s" % __version__)
        return
    basicConfig(level=10 if args.debug else 20, format="[%(levelname)s] %(message)s")

    #print(args)
    if args.vocabfile is None:
        words = default_wordle_vocab(args.answer_difficulty)
        vocabname = args.vocabname
        if vocabname is None:
            vocabname = ("wordle_lev%s" % args.answer_difficulty)
        #print(vocabname)
    else:
        words = _read_vocabfile(args.vocabfile)
        vocabname = args.vocabname
        if vocabname is None:
            vocabname = re.sub(r"\..*$", "", os.path.basename(args.vocabfile))

    #print(words)
    if args.play:
        while True:
            play(words, vocabname=vocabname)
            while True:
                ans = input("One more game? (y/n) > ")
                ans = ans.strip().lower()[0:1]
                if ans in ("y", "n"):
                    break
            if ans == "n":
                print("Thank you!")
                return

    if args.backend == "sqlite":
        if args.inmemory:
            logger.warning("`--inmemory` only applicable with `-b approx`")
        ai = WordleAISQLite(vocabname, words, dbfile=args.sqlitefile, resetup=args.resetup,
                            decision_metric=args.decision_metric, candidate_weight=args.candidate_weight, strength=args.ai_strength,
                            use_cpp=(not args.no_cpp), cpp_recompile=args.cpp_recompile, cpp_compiler=args.cpp_compiler)
        logger.info("SQLite database: '%s', vocabname: '%s'", ai.dbfile, ai.vocabname)
    elif args.backend == "approx":
        ai = WordleAIApprox(vocabname, words, dbfile=args.sqlitefile, inmemory=args.inmemory, resetup=args.resetup,
                            word_pair_limit=args.word_pair_limit, candidate_samplesize=args.candidate_samplesize,
                            decision_metric=args.decision_metric, candidate_weight=args.candidate_weight, strength=args.ai_strength)
        logger.info("SQLite database: '%s', word pair limit: %d, answer word sample size: %d, vocabname: '%s'",
                    ai.dbfile, ai.word_pair_limit, ai.candidate_samplesize, ai.vocabname)
    elif args.backend == "bq":
        from .bigquery import WordleAIBigquery
        ai = WordleAIBigquery(vocabname, words, resetup=args.resetup,
                              credential_jsonfile=args.bq_credential, project=args.bq_project,
                              location=args.bq_location, partition_size=args.partition_size,
                              decision_metric=args.decision_metric, candidate_weight=args.candidate_weight, strength=args.ai_strength)
        logger.info("GCP project: '%s', location: '%s', vocabname: '%s'", ai.project, ai.location, ai.vocabname)
    elif args.backend == "random":
        ai = WordleAI(vocabname, words)
    else:
        raise ValueError("Backend not supported '%s'" % args.backend)

    if args.challenge:
        while True:
            challenge(ai, max_round=args.max_round, visible=args.visible, alternate=args.alternate,
                      ai_first=args.ai_first, continue_after_result=args.continue_after_result)
            while True:
                ans = input("One more game? (y/n) > ")
                ans = ans.strip().lower()[0:1]
                if ans in ("y", "n"):
                    break
            if ans == "n":
                print("Thank you!")
                return
    else:
        return interactive(ai, num_suggest=args.num_suggest, default_criterion=args.suggest_criterion)


# -*- coding: utf-8 -*-

import random
from typing import Type
from .utils import wordle_judge, encode_judgement, WordEvaluation, _dedup, _read_vocabfile

class WordleAI:
    """
    Base Wordle AI class.
    
    - Keeps words as the instance variable
    - Evaluation result is random
    - Pick a word in the candidate list randomly

    Args:
        vocabname (str):
            Name of vocaburary
        words (str or list or dict):
            If str, the path to a vocabulary file
            If list, the list of words
            If dict, mapping from word to the weight
    """
    def __init__(self, vocabname: str, words: str or list or dict=None, **kwargs):
        self.vocabname = vocabname
        self._vocabnames = [vocabname]  # no storage of other vocabs
        assert words is not None, "`words` must be supplied to setup the vocab '{}'".format(vocabname)
        if isinstance(words, dict):
            # maps a word to float
            self._words = words.copy()
        elif isinstance(words, list):
            # list of words with equal weight
            self._words = {w:1.0 for w in words}
        elif isinstance(words, str):
            # file path
            self._words = _read_vocabfile(words)
        else:
            raise TypeError("Unsupported type 'words': '{}'".format(type(words)))
        assert len(self._words) > 0, "Empty vocab is not allowed"
        wordlens = set(len(w) for w in self.words)
        assert len(wordlens) == 1, "word length must be equal, but '{}'".format(wordlens)

        self._info = []  # infomation of the judge result
        self._nonanswer_words = set([])  # words that cannot become an answer

    @property
    def name(self)-> str:
        return "Wordle AI (random)"

    @property
    def vocabnames(self)-> list:
        """Available vocab names"""
        return self._vocabnames if hasattr(self, "_vocabnames") else []

    @property
    def words(self)-> list:
        """All words that can be inputted"""
        return list(self._words.keys())

    @property
    def candidates(self)-> list:
        """Subset of answer words filtered by given information"""
        candidates = [w for w in self.words if w not in self.nonanswer_words]
        def _keep(candidate):
            for input_word, encoded_result in self.info:
                if wordle_judge(input_word, candidate) != int(encoded_result):
                    return False
            return True
        candidates = [c for c in candidates if _keep(c)]
        return candidates
        #return self._candidates

    @property
    def info(self)-> list:
        """List of (input_word, decoded_result)"""
        return self._info

    def clear_info(self):
        self.info.clear()

    @property
    def nonanswer_words(self)-> set:
        """Set of words that cannot become an answer"""
        return self._nonanswer_words

    def remove_from_answers(self, excluded_words: list):
        self._nonanswer_words |= set(excluded_words)

    # def set_candidates(self, candidates: list=None):
    #     """
    #     Set the candidate words.
    #     If candidates is none, then reset to the all answer words.
    #     """
    #     if candidates is not None:
    #         self._candidates = _dedup(candidates)
    #     else:
    #         self._candidates = self.words.copy()

    def evaluate(self, top_k: int=20, criterion: str="mean_entropy")-> list:
        """
        Evaluate input words and return the top ones in accordance with the given criterion
        """
        # this class picks a random candidate word
        n = min(top_k, len(self.candidates))
        results = {c: WordEvaluation(c, 1, 1, 1, 1) for c in random.sample(self.candidates, n)}
        if len(results) >= top_k:
            return list(results.values())
        # if there are less than top_k words in candidates add some answer words
        for w in self.words:
            if w in results:
                continue
            results[w] = WordEvaluation(w, 1, 1, 1, 0)
            if len(results) >= top_k:
                break
        results = list(results.values())
        results.sort(key=lambda row: -row[-1])
        return results

    def update(self, input_word: str, judge_result: int or str):
        """
        Update information on a judge result.

        Judge result is decoded, i.e. human-interpretable one such as '22001'
        """
        # self.set_candidates([c for c in self.candidates
        #                      if wordle_judge(input_word, c) == encode_judgement(int(judge_result))])
        self.info.append((input_word, encode_judgement(int(judge_result))))

    def pick_word(self, criterion: str="mean_entropy")-> str:
        """Pick an input word"""
        if len(self.candidates) > 0:
            return random.choice(self.candidates)
        print("Warning: There is no answer candidates remaining")
        return random.choice(self.words)
    
    def choose_answer_word(self):
        """Randomly choose an answer word in accordance with the given weight"""
        vals = []
        weights = []
        for w, p in self._words.items():
            if p > 0:
                vals.append(w)
                weights.append(p)
        assert len(vals) > 0, "There is no word with positive weight"
        return random.choices(vals, weights, k=1)[0]


# -*- coding: utf-8 -*-

"""
Bigquery backend.

Tables are stored in the following structure:

{project}.{vocabname}.words   : contains all words
{project}.{vocabname}.judges  : contains judge results for all word pairs

i.e. {vocabname} corresponds to the dataset name
"""

import time
import math
import sys
import random
from logging import getLogger
logger = getLogger(__name__)

try:
    from google.cloud import bigquery
    from google.cloud.exceptions import NotFound, Forbidden
    #from google.cloud.bigquery import dbapi
except (ModuleNotFoundError, ImportError) as e:
    raise RuntimeError("Import failed: '{}'. Please install bigquery module by `pip install google-cloud-bigquery`".format(e))

from .utils import _timereport, WordEvaluation, _read_vocabfile, _dedup
from .sqlite import WordleAISQLite

def _make_client(credential_jsonfile: str=None, **kwargs):
    if credential_jsonfile is None:
        return bigquery.client.Client(**kwargs)
    else:
        return bigquery.client.Client.from_service_account_json(credential_jsonfile, **kwargs)

def _create_dataset(client: bigquery.Client, datasetname: str, location: str=None):
    dataset = bigquery.dataset.Dataset(client.dataset(datasetname))
    if location is not None:
        dataset.location = location
    return client.create_dataset(dataset, exists_ok=True)

def _setup(client: bigquery.Client, vocabname: str, words: list or dict, project: str=None, location: str="US", partition_size: int=200):
    assert len(words) == len(set(words)), "input_words must be unique"
    wordlens = set(len(w) for w in words)
    assert len(wordlens) == 1, "word length must be equal, but '{}'".format(wordlens)

    if project is None:
        project = client.project
    if location is None:
        location = client.location
    logger.info("Bigquery project: '%s', location: '%s'", project, location)
    
    logger.info("Creating dataset '%s'", vocabname)
    dataset = _create_dataset(client, vocabname, location=location)
    #print(dataset)
    #print(vars(dataset))

    schema = [bigquery.SchemaField("word", "STRING", mode="REQUIRED"),
              bigquery.SchemaField("weight", "FLOAT", mode="REQUIRED"),
              bigquery.SchemaField("partid", "INTEGER", mode="REQUIRED")]
    tableid = "{}.{}.words".format(project, vocabname)
    table = bigquery.Table(tableid, schema=schema)

    logger.info("Deleting table 'words' (if exists)")
    client.delete_table(table, not_found_ok=True)    
    logger.info("Creating table 'words'")
    table = client.create_table(table)
    # we assign partition ID to each word
    #client.query('CREATE OR REPLACE TABLE {project}.{dataset}.words (word STRING, partid INTEGER)'.format(project=project, dataset=vocabname)).result()
    rows = (
        [(w, p, i % partition_size) for i, (w, p) in enumerate(words.items())] if isinstance(words, dict) else
        [(w, 1, i % partition_size) for i, w in enumerate(words)] if isinstance(words, list) else
        None
    )
    if rows is None:
        raise TypeError("Unsupported type of `words`, '{}'".format(type(words)))
    #    rows = [(w, i % partition_size) for i, w in enumerate(words)]
    min_partid = 0
    max_partid = max(row[-1] for row in rows)
    # it can take some time until table is found
    time.sleep(20)
    logger.info("Inserting data to table 'words'")
    for i in range(100):
        try:
            client.insert_rows(table, rows)
            logger.info("Inserted data to table 'words'")
            break
        except (NotFound, Forbidden) as e:
            logger.info("Insert to the table 'words' failed (trial %s, %s)", i+1, e)
            time.sleep(15)
    #client.query('DELETE FROM `{project}.{dataset}.words`'.format(project=project, dataset=vocabname)).result()
    

    #client.query('INSERT INTO `{project}.{dataset}.words VALUES (?,?)')
    # time.sleep(15)  # there is some lag before we can insert rows
    # for i in range(20):
    #     try:
    #         client.insert_rows(table, rows)
    #         break
    #     except Exception as e:
    #         print(type(e))
    #         logger.info("Insert failed ('%s') at the try %d", e, i+1)
    #         time.sleep(10)

    # create UDF
    js = """
    // (input_word, answer_word) --> int

    var n_letter = input_word.length;
    // check exact match
    var exact_match = Array(n_letter);
    for (let i=0; i<n_letter; i++) {
        exact_match[i] = (input_word[i]==answer_word[i]);
    }

    // letter count except for the exact math
    var letter_count = {}
    for (let i=0; i<n_letter; i++) {
        if (exact_match[i]) continue;
        if (answer_word[i] in letter_count) {
            letter_count[answer_word[i]]++;
        } else {
            letter_count[answer_word[i]] = 1;
        }
    }

    // define partial match
    var partial_match = Array.from({length: n_letter}, i => false);
    for (let i=0; i<n_letter; i++) {
        if (exact_match[i]) continue;
        if (input_word[i] in letter_count) {
            partial_match[i] = true;
            letter_count[input_word[i]]--;
        } 
    }

    var out = 0;
    var power = 1;
    for (let i=0; i<n_letter; i++) {
        if (exact_match[n_letter-1-i]) {
            out += (power * 2);
        } else if (partial_match[n_letter-1-i]) {
            out += power;
        }
        power *= 3;
    }  
    return out;  
    """
    q = '''
    CREATE FUNCTION IF NOT EXISTS `{project}.{dataset}.WordleJudge` (input_word STRING, answer_word STRING)
      RETURNS INTEGER
      LANGUAGE js AS """{js}""";
    '''.format(project=project, dataset=vocabname, js=js)    
    logger.info("Creating udf 'WordleJudge'")    
    client.query(q).result()

    with _timereport("precomputing wordle judges"):
        # client.query('''
        # DROP TABLE IF EXISTS {project}.{dataset}.judges
        # '''.format(project=project, dataset=vocabname)).result()
        
        q = '''
        CREATE OR REPLACE TABLE {project}.{dataset}.judges
          PARTITION BY RANGE_BUCKET(answer_word_partid, GENERATE_ARRAY({min_partid}, {max_partid}))
          CLUSTER BY input_word, judge
        AS
          SELECT
            a.word AS input_word,
            b.word AS answer_word,
            {project}.{dataset}.WordleJudge(a.word, b.word) AS judge,
            b.partid AS answer_word_partid
          FROM
            {project}.{dataset}.words AS a, {project}.{dataset}.words AS b
        '''.format(project=project, dataset=vocabname, min_partid=min_partid, max_partid=max_partid+1)
        job = client.query(q)
        job.result()
        logger.info("Total bytes processed: %.2f GB, total bytes billed: %.2f GB",
                    1.0*job.total_bytes_processed/1073741824, 1.0*job.total_bytes_billed/1073741824)

# def _ensure_word_weight_column(client: bigquery.Client, vocabname: str, project: str=None):
#     """If weight column is missing in the words table, add it with a constant 1"""
#     if project is None:
#         project = client.project
#     q = """
#     SELECT column_name FROM {project}.{dataset}.information_schema.columns
#     WHERE
#       table_name = '{project}.{dataset}.words'
#       AND lower(column_name) = 'weight'
#     LIMIT 1
#     """.format(project=project, dataset=vocabname)
#     job = client.query(q)
#     res = list(job.result())
#     if len(res) == 0:  # weight column not exist
#         job = client.query('ALTER TABLE {project}.{dataset}.words ADD COLUMN weight FLOAT'.format(project=project, dataset=vocabname))
#         job.result()
#         logger.info('Added column `%s.%s.words.weight`', project, vocabname)
#         job = client.query('UPDATE "{project}.{dataset}.words SET weight = 1.0'.format(project=project, dataset=vocabname))
#         logger.info('Filled `%s.%s.words.weight` with ones', project, vocabname)
        
def _evaluate(client: bigquery.Client, vocabname: str, project: str,
              top_k: int=20, criterion: str="mean_entropy", candidates: list=None)-> list:
    # find the number of all words and compare with the number of candidates
    # if they are the same, then we do not need to filter answer_word
    job = client.query('SELECT count(*) FROM {project}.{dataset}.words'.format(project=project, dataset=vocabname))
    rows = job.result()    
    n_words = next(rows)[0]

    if candidates is None or len(candidates) >= n_words:  # all words are in the candidates
        params = None
        answer_filter = ""
    else:
        candidate_set = set(candidates)
        params = [bigquery.ScalarQueryParameter(None, "STRING", c) for c in candidate_set]
        answer_filter = """
        WHERE
          answer_word_partid IN (SELECT partid FROM {project}.{dataset}.words WHERE word IN ({placeholder}))
          AND
          answer_word IN ({placeholder}) 
        """.format(project=project, dataset=vocabname, placeholder=",".join("?" * len(params)))
        params = params + params
        # answer_filter = """
        # INNER JOIN
        #   {project}.{dataset}.words AS b
        #   ON
        #     b.word IN ({placeholder})
        #     AND a.answer_word_partid = b.partid
        #     AND a.answer_word = b.word          
        # """.format(project=project, dataset=vocabname, placeholder=",".join("?" * len(params)))
        # params = params
        # # join version costs more we go with the IN version above
    q = """
    with tmp AS (
      SELECT
        input_word,
        judge,
        count(*) AS n,
        log(count(*), 2) AS entropy
      FROM
        {project}.{dataset}.judges
      {answerfilter}
      GROUP BY
        input_word, judge
    )
    SELECT
      input_word,
      max(n) AS max_n,
      1.0 * sum(n*n) / sum(n) AS mean_n,
      sum(n*entropy) / sum(n) AS mean_entropy
    FROM
      tmp
    GROUP BY
      input_word
    """.format(answerfilter=answer_filter, project=project, dataset=vocabname)
    #print(q)
    #print(params)
    if params is None:
        job = client.query(q)
    else:
        job_config = bigquery.QueryJobConfig(query_parameters=params)
        job = client.query(q, job_config=job_config)
    rows = job.result()
    #print(job.total_bytes_processed)
    #print(job.total_bytes_billed)
    logger.info("Total bytes processed: %.2f GB, total bytes billed: %.2f GB",
                1.0*job.total_bytes_processed/1073741824, 1.0*job.total_bytes_billed/1073741824)
    # with _connect(client) as conn:
    #     c = conn.cursor()
    #     if params is None:
    #         c.execute(q)
    #     else:
    #         c.execute(q, params)
    candidate_set = None if candidates is None else set(candidates)
    out = [tuple(row) for row in rows]
    out = [row + (1 if candidate_set is None else int(row[0] in candidate_set),) for row in out]
    out = [WordEvaluation(*row) for row in out]
    out.sort(key=lambda row: (getattr(row, criterion), -row.is_candidate))
    return out[:top_k]

def _vocabnames(client: bigquery.Client, project: str)-> list:
    # return datasets that includes words and judges table
    datasets = client.list_datasets(project=project)
    out = []
    for d in datasets:
        tables = client.list_tables(d)
        tableids = set(t.table_id for t in tables)
        if "words" in tableids and "judges" in tableids:
            out.append(d.dataset_id)
    return out

def _words(client: bigquery.Client, vocabname: str, project: str)-> list:
    job = client.query('SELECT word FROM `{project}.{dataset}.words`'.format(project=project, dataset=vocabname))
    rows = job.result()
    words = [row[0] for row in rows]
    return words

def _choose_word_with_weight(client: bigquery.Client, vocabname: str, project: str)-> str:
    # The query below uses the fact that
    # Prob{ u_i^(1/w_i) > u_j^(1/w_j) } = w_i / (w_i + w_j),
    #   where u_i, u_j ~ uniform(0, 1).
    #
    # References: 
    #   https://stackoverflow.com/questions/1398113/how-to-select-one-row-randomly-taking-into-account-a-weight
    #   http://utopia.duth.gr/~pefraimi/research/data/2007EncOfAlg.pdf
    # 
    # On Bigquery, rand() is uniform on [0, 1)
    # We use 1 - rand() to make it to (0, 1] to avoid log(0).
    #   note. log(1) is okay because it is zero.
    q = """
    SELECT
        word,
        -log(1 - rand()) / weight AS priority
    FROM
        {project}.{dataset}.words
    WHERE
        weight > 0
    ORDER BY priority
    LIMIT 1
    """.format(project=project, dataset=vocabname)
    job = client.query(q)
    rows = job.result()
    return next(rows)[0]

def _weight_defined(client: bigquery.Client, vocabname: str, project: str=None)-> bool:
    """If weight column is missing in the words table, add it with a constant 1"""
    if project is None:
        project = client.project
    q = """
    SELECT * FROM {project}.{dataset}.words LIMIT 1
    """.format(project=project, dataset=vocabname)
    job = client.query(q)
    res = next(job.result())
    return ("weight" in res.keys())


class WordleAIBigquery(WordleAISQLite):
    """
    Wordle AI with SQLite backend

    Vocab information is stored in {vocabname}_words and {vocabname}_judges

    Args:
        vocabname (str):
            Name of vocaburary
        words (str or list or dict): 
            If str, the path to a vocabulary file
            If list, the list of words
            If dict, mapping from word to the weight
            Can be omitted if the vocabname is already in the database and resetup=False
        credential_jsonfile (str):
            Path to the service accound credential file
            If not supplied, authenticate with no file
        project (str):
            GCP project id to use
            If not supplied, use the default project of the client
        location (str):
            Location of the bigquery tables
        partition_size (int):
            Partition size of judges table
            The rows of judges table are partitioned by a set of answer_word
            both for cost reduction and computaion speed

        decision_metric (str):
            The criteria to pick a word
            Either 'max_n', 'mean_n', of 'mean_entropy'
        candidate_weight (float):
            The weight added to the answer candidate word when picking a word
        strength (float):
            AI strength in [0, 10]

        resetup (bool):
            Setup again if the vocabname already exists        
    """
    def __init__(self, vocabname: str, words: str or list or dict=None,
                 credential_jsonfile: str=None, project: str=None, location: str="US", partition_size: int=200,
                 decision_metric: str="mean_entropy", candidate_weight: float=0.3, strength: float=6,
                 resetup: bool=False, **kwargs):
        self.client = _make_client(credential_jsonfile, project=project, location=location)
        self.project = self.client.project
        self.location = self.client.location
        logger.info("GCP project: '%s', location: '%s'", self.project, self.location)

        self.vocabname = vocabname
        self.decision_metric = decision_metric
        self.candidate_weight = candidate_weight
        self.strength = min(max(strength, 0), 10)  # clip to [0, 10]
        # strength is linearly converted to the power of noise: 0 -> +5, 10 -> -5
        # larger noise, close to random decision
        self.decision_noise = math.pow(10, 5-self.strength)

        if resetup or vocabname not in self.vocabnames:
            assert words is not None, "`words` must be supplied to setup the vocab '{}'".format(vocabname)
            assert words is not None, "`words` must be supplied to setup the vocab '{}'".format(vocabname)
            _words = (
                words if isinstance(words, dict) else
                {w:1.0 for w in words} if isinstance(words, list) else
                _read_vocabfile(words) if isinstance(words, str) else
                None
            )
            if _words is None:
                raise TypeError("Unsupported type 'words': '{}'".format(type(words)))
            with _timereport("Setup tables for vocabname '%s'" % vocabname):
                _setup(client=self.client, vocabname=self.vocabname, words=words,
                       project=self.project, location=self.location, partition_size=partition_size)
        # else:
        #     _ensure_word_weight_column(client=self.client, vocabname=self.vocabname, project=self.project)

        #self.set_candidates()
        self._info = []                  # infomation of the judge results
        self._nonanswer_words = set([])  # words that cannot become an answer

    @property
    def name(self)-> str:
        return "Wordle AI (bigquery backend)"

    @property
    def vocabnames(self)-> list:
        """Available vocab names"""
        return _vocabnames(client=self.client, project=self.project)

    @property
    def words(self)-> set:
        """All words that can be inputted"""
        return _words(client=self.client, vocabname=self.vocabname, project=self.project)
    
    def evaluate(self, top_k: int=20, criterion: str="mean_entropy")-> list:
        """
        Evaluate input words and return the top ones in accordance with the given criterion
        """
        return _evaluate(self.client, self.vocabname, self.project,
                         top_k=top_k, criterion=criterion, candidates=self.candidates)

    def choose_answer_word(self)-> str:
        """Randomly choose an answer word in accordance with the given weight"""
        if not _weight_defined(self.client, self.vocabname, self.project):
            print("Word weight is not defined. Please call `WordleAIBigquery` with `resetup=True` next time", file=sys.stderr)
            return random.choice(self.words)
        return _choose_word_with_weight(self.client, self.vocabname, self.project)


# -*- coding: utf-8 -*-

"""
SQLite backend.

Tables are named by the following convention:

{vocabname}_words   : contains all words
{vocabname}_judges  : contains judge results for all word pairs
"""

import os
import sys
import sqlite3
import math
import random
from contextlib import contextmanager
from logging import getLogger
logger = getLogger(__name__)

from .utils import all_wordle_judges, _timereport, _dedup, WordEvaluation, wordle_judge, encode_judgement, _read_vocabfile
from .base import WordleAI


def _setup(dbfile: str, vocabname: str, words: list or dict, use_cpp: bool=True, recompile: bool=False, compiler: str=None):
    assert len(words) == len(set(words)), "input_words must be unique"
    wordlens = set(len(w) for w in words)
    assert len(wordlens) == 1, "word length must be equal, but '{}'".format(wordlens)
    with sqlite3.connect(dbfile) as conn:
        c = conn.cursor()
        c.execute("PRAGMA journal_mode=OFF")  # disable rollback to save time        
        
        c.execute('DROP TABLE IF EXISTS "{name}_words"'.format(name=vocabname))
        c.execute('CREATE TABLE "{name}_words" (word TEXT PRIMARY KEY, weight FLOAT)'.format(name=vocabname))
        if isinstance(words, dict):
            params = words.items()
        elif isinstance(words, list):
            params = [(w, 1) for w in words]
        else:
            raise TypeError("Unsupported type of `words`, '{}'".format(type(words)))
        c.executemany('INSERT INTO "{name}_words" VALUES (?,?)'.format(name=vocabname), params)
        c.execute('CREATE INDEX "{name}_words_idx" ON "{name}_words" (word)'.format(name=vocabname))

        with _timereport("Precomputing wordle judges"):
            c.execute('DROP TABLE IF EXISTS "{name}_judges"'.format(name=vocabname))
            c.execute('CREATE TABLE "{name}_judges" (input_word TEXT, answer_word TEXT, judge INT)'.format(name=vocabname))
            params = all_wordle_judges(words, use_cpp=use_cpp, recompile=recompile, compiler=compiler)
            c.executemany('INSERT INTO "{name}_judges" VALUES (?,?,?)'.format(name=vocabname), params)

        with _timereport("Creating indices"):
            c.execute('CREATE INDEX "{name}_judge_idx" ON "{name}_judges" (input_word, judge)'.format(name=vocabname))
            c.execute('CREATE INDEX "{name}_judge_idx2" ON "{name}_judges" (answer_word)'.format(name=vocabname))
        conn.commit()

# def _ensure_word_weight_column(dbfile: str, vocabname: str):
#     """If weight column is missing in the words table, add it with a constant 1"""
#     with sqlite3.connect(dbfile) as conn:
#         c = conn.cursor()
#         # check the existing column
#         c.execute('SELECT * FROM "{name}_words" LIMIT 1'.format(name=vocabname))
#         exist = any(col[0].lower() == "weight" for col in c.description)
#         if not exist:
#             c.execute('ALTER TABLE "{name}_words" ADD weight FLOAT'.format(name=vocabname))
#             logger.info('Added column `"%s_words".weight`', vocabname)
#             c.execute('UPDATE "{name}_words_approx" SET weight = 1.0'.format(name=vocabname))
#             logger.info('Filled `"%s_words_approx".weight` with ones', vocabname)
#         conn.commit()

def _evaluate(dbfile: str, vocabname: str, top_k: int=20, criterion: str="mean_entropy", candidates: list=None)-> list:
    with sqlite3.connect(dbfile) as conn:
        conn.create_function("log2", 1, math.log2)
        c = conn.cursor()
        # find the number of all words and compare with the number of candidates
        # if they are the same, then we do not need to filter answer_word
        c.execute('SELECT count(*) FROM "{name}_words"'.format(name=vocabname))
        n_words = c.fetchone()[0]

        if candidates is None or len(candidates) >= n_words:  # all words are in the candidates
            params = None
            answer_filter = ""
        else:
            candidate_set = set(candidates)
            params = tuple(candidate_set)
            answer_filter = "WHERE answer_word IN (%s)" % ",".join("?" * len(params))

        q = """
        with tmp AS (
          SELECT
            input_word,
            judge,
            count(*) AS n,
            log2(count(*)) AS entropy
          FROM
            "{name}_judges"
          {answerfilter}
          GROUP BY
            input_word, judge
        )
        SELECT
          input_word,
          max(n) AS max_n,
          1.0 * sum(n*n) / sum(n) AS mean_n,
          sum(n*entropy) / sum(n) AS mean_entropy
        FROM
          tmp
        GROUP BY
          input_word
        """.format(answerfilter=answer_filter, name=vocabname)
        #print(q)
        if params is None:
            c.execute(q)
        else:
            c.execute(q, params)

        candidate_set = None if candidates is None else set(candidates)
        out = [row + (1 if candidate_set is None else int(row[0] in candidate_set),) for row in c]
    out = [WordEvaluation(*row) for row in out]
    out.sort(key=lambda row: (getattr(row, criterion), -row.is_candidate))
    return out[:top_k]

def _vocabnames(dbfile: str)-> list:
    with sqlite3.connect(dbfile) as conn:
        c = conn.cursor()
        c.execute("SELECT name FROM sqlite_master")
        tables = [row[0] for row in c]
        t1 = [t[:-6] for t in tables if t.endswith("_words")]
        t2 = [t[:-7] for t in tables if t.endswith("_judges")]
        out = list(set(t1) & set(t2))  # we need both _words and _judges tables
    return out

def _words(dbfile: str, vocabname: str)-> list:
    with sqlite3.connect(dbfile) as conn:
        c = conn.cursor()
        c.execute('SELECT * FROM "{name}_words"'.format(name=vocabname))
        words = [row[0] for row in c]
    return words

def _choose_word_with_weight(dbfile: str, vocabname: str)-> str:
    with sqlite3.connect(dbfile) as conn:
        sqlite3.enable_callback_tracebacks(True)
        conn.create_function("log", 1, math.log)
        c = conn.cursor()
        # The query below uses the fact that
        # Prob{ u_i^(1/w_i) > u_j^(1/w_j) } = w_i / (w_i + w_j),
        #   where u_i, u_j ~ uniform(0, 1).
        #
        # References: 
        #   https://stackoverflow.com/questions/1398113/how-to-select-one-row-randomly-taking-into-account-a-weight
        #   http://utopia.duth.gr/~pefraimi/research/data/2007EncOfAlg.pdf
        # 
        # On SQLite, random() generate an integer from [-9223372036854775808, +9223372036854775808]
        # Remove sign and take mod n to make them roughly uniform on [0, n-1]
        # Then plus 0.5 to avoid zero, i.e. uniform on [0.5, ..., n-0.5]
        # Dividing by n, uniform on [0.5/n, ... 1-0.5/n]
        # With n large enough, the retult is roughly uniform on (0, 1)
        q = """
        SELECT
          word,
          weight / log((abs(random()) % 100000 + 0.5) / 100000.0) AS priority
        FROM
          "{name}_words"
        WHERE
          weight > 0
        ORDER BY priority
        LIMIT 1
        """.format(name=vocabname)
        c.execute(q)
        ans = c.fetchall()
    return ans[0][0]

def _weight_defined(dbfile: str, vocabname: str)-> bool:
    with sqlite3.connect(dbfile) as conn:
        c = conn.cursor()
        # check the existing column
        c.execute('SELECT * FROM "{name}_words" LIMIT 1'.format(name=vocabname))
        for col in c.description:
            if col[0].lower() == "weight":
                return True
    return False


class WordleAISQLite(WordleAI):
    """
    Wordle AI with SQLite backend

    Vocab information is stored in {vocabname}_words and {vocabname}_judges

    Args:
        vocabname (str):
            Name of vocaburary
        words (str or list or dict): 
            If str, the path to a vocabulary file
            If list, the list of words
            If dict, mapping from word to the weight
            Can be omitted if the vocabname is already in the database and resetup=False
        dbfile (str):
            SQLite database file
            If not supplied, use environment variable `WORDLEAISQL_DBFILE` if exists,
            otherwise './wordleai.db' in the current directory is used

        decision_metric (str):
            The criteria to pick a word
            Either 'max_n', 'mean_n', of 'mean_entropy'
        candidate_weight (float):
            The weight added to the answer candidate word when picking a word
        strength (float):
            AI strength in [0, 10]

        use_cpp (bool):
            Use C++ code to precompute wodle judgements when available
        cpp_recompile (bool):
            Compile the C++ code again if the source code has no change
        cpp_compiler (str):
            Command name of the C++ compiler. If None, 'g++' and 'clang++' are searched

        resetup (bool):
            Setup again if the vocabname already exists
    """
    def __init__(self, vocabname: str, words: str or list or dict=None, dbfile: str=None,
                 decision_metric: str="mean_entropy", candidate_weight: float=0.3, strength: float=6,
                 use_cpp: bool=True, cpp_recompile: bool=False, cpp_compiler: str=None, resetup: bool=False, **kwargs):
        if dbfile is None:
            dbfile = os.environ.get("WORDLEAISQL_DBFILE")
            if dbfile is None:
                dbfile = "./wordleai.db"
        os.makedirs(os.path.dirname(os.path.abspath(dbfile)), exist_ok=True)
        self.dbfile = dbfile
        logger.info("SQLite database: '%s'", self.dbfile)
        self.vocabname = vocabname
        self.decision_metric = decision_metric
        self.candidate_weight = candidate_weight
        self.strength = min(max(strength, 0), 10)  # clip to [0, 10]
        # strength is linearly converted to the power of noise: 0 -> +5, 10 -> -5
        # larger noise, close to random decision
        self.decision_noise = math.pow(10, 5-self.strength)

        if resetup or (vocabname not in self.vocabnames):
            assert words is not None, "`words` must be supplied to setup the vocab '{}'".format(vocabname)
            _words = (
                words if isinstance(words, dict) else
                {w:1.0 for w in words} if isinstance(words, list) else
                _read_vocabfile(words) if isinstance(words, str) else
                None
            )
            if _words is None:
                raise TypeError("Unsupported type 'words': '{}'".format(type(words)))
            with _timereport("Setup tables for vocabname '%s'" % vocabname):
                _setup(dbfile=dbfile, vocabname=vocabname, words=_words, use_cpp=use_cpp, recompile=cpp_recompile, compiler=cpp_compiler)
        # else:
        #     _ensure_word_weight_column(dbfile, vocabname)  # make sure previously created words table has the weight column

        self._info = []                  # infomation of the judge results
        self._nonanswer_words = set([])  # words that cannot become an answer
        #self.set_candidates()

    @property
    def name(self)-> str:
        return "Wordle AI (SQLite backend)"

    @property
    def vocabnames(self)-> list:
        """Available vocab names"""
        return _vocabnames(self.dbfile)

    @property
    def words(self)-> list:
        """All words that can be inputted"""
        return _words(self.dbfile, self.vocabname)

    def evaluate(self, top_k: int=20, criterion: str="mean_entropy")-> list:
        """
        Evaluate input words and return the top ones in accordance with the given criterion
        """
        return _evaluate(self.dbfile, self.vocabname, top_k=top_k, criterion=criterion, candidates=self.candidates)
    
    def pick_word(self):
        num_remain = len(self.candidates)
        #print(count, candidates)
        if num_remain == 1:
            return self.candidates[0]
        elif num_remain == 0:
            print("Warning: No candidates left. This is a random choice")
            return random.choice(self.words)

        results = self.evaluate(top_k=10000, criterion=self.decision_metric)
        #print(results[:10], len(results))        
        words = [row.input_word for row in results]
        scores = [getattr(row, self.decision_metric) for row in results]  # score of eadch word, the smaller the better
        if self.decision_metric in ("mean_n", "max_n"):
            # we take log of the score to adjust for the scale
            # add 1p just in case to avoid the zero error
            scores = [math.log1p(s) for s in scores]
        
        # Flip the sign and adjust for the candidates
        for i, row in enumerate(results):
            scores[i] = row.is_candidate * self.candidate_weight - scores[i]
        # Subtract the maximum to avoid overflow
        maxscore = max(scores)
        scores = [s - maxscore for s in scores]
        # Add randomness
        weights = [math.exp(s / self.decision_noise) for s in scores]
        
        out = random.choices(words, weights=weights, k=1)
        return out[0]

    def choose_answer_word(self)-> str:
        """Randomly choose an answer word in accordance with the given weight"""
        if not _weight_defined(self.dbfile, self.vocabname):
            print("Word weight is not defined. Please call `WordleAISQLite` with `resetup=True` next time", file=sys.stderr)
            return random.choice(self.words)
        return _choose_word_with_weight(self.dbfile, self.vocabname)


# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import os
readmefile = os.path.join(os.path.dirname(__file__), "README.md")
with open(readmefile) as f:
    readme = f.read()

from wordleaisql import __version__

setup(
    name='wordleaisql',
    version=__version__,
    description='Wordle AI with SQL Backend',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/kota7/wordleai-sql',
    packages=['wordleaisql'],
    install_requires=['tqdm'],
    #test_require=[],
    package_data={"wordleaisql": ["wordle-judge-all.cpp", "vocab/*.txt.gz"]},
    entry_points={'console_scripts': ['wordleai-sql=wordleaisql.api:main']},
    
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        #'Programming Language :: Python :: 2.6',
        #'Programming Language :: Python :: 2.7',
        #'Programming Language :: Python :: 3.3',
        #'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    test_suite='tests'
)



# -*- coding: utf-8 -*-

import os
import gzip
import sys
import itertools
import hashlib
import subprocess
from collections import Counter, namedtuple
from contextlib import contextmanager
from datetime import datetime
from tempfile import TemporaryDirectory
from logging import getLogger
logger = getLogger(__name__)

from tqdm import tqdm

def _dedup(x: list)-> list: 
    return list(dict.fromkeys(x))

def wordle_judge(input_word: str, answer_word: str)-> int:
    """
    Judge input word based on the wordle rule
    
    We assume input_word and answer_word are of the same length (but no check is conducted)

    Judge results are defined as a sequence of {0, 1, 2}, where
      2: exact, 1: partial, 0: none
    Return value is the base-10 integer value that represents this result interpreted as an integer of base 3.

    e.g. 22001 --> 2 * 3^4 + 2 * 3*3 + 0 * 3^2 + 0 * 3^2 + 1 * 3^0 = 181
    """
    
    exactmatch = [a==b for a, b in zip(input_word, answer_word)]
    lettercount = Counter(b for b, m in zip(answer_word, exactmatch) if not m)
    partialmatch = [False] * len(input_word)
    for i, (a, m) in enumerate(zip(input_word, exactmatch)):
        if m: continue
        if lettercount.get(a, 0) > 0:
            lettercount[a] -= 1
            partialmatch[i] = True
    out = 0
    power = 1
    for x, y in zip(reversed(exactmatch), reversed(partialmatch)):
        if x:
            out += power*2
        elif y:
            out += power
        power *= 3
    return out

def decode_judgement(number: int or str)-> int:
    # convert to human-friendly integer
    number = int(number)
    out = 0
    power = 1
    while number > 0:
        out += power*(number % 3)
        number = int(number / 3)
        power *= 10
    return out

def encode_judgement(number: int)-> int:
    # convert to expression system 
    if type(number) != int:
        number = int(number)
    out = 0
    power = 1
    while number > 0:
        out += power*(number % 10)
        number = int(number / 10)
        power *= 3
    return out


# Evaluation of input word
WordEvaluation = namedtuple("WordEvaluation", "input_word max_n mean_n mean_entropy is_candidate")
def show_word_evaluations(x: list):
    # evaluation result is a list of dict, with the same keys repeated
    if len(x) == 0:
        print("No data.")
        return
    keys = ("input_word", "max_n", "mean_n", "mean_entropy", "is_candidate")
    rowfmt = "%12s  %12s  %12s  %12s  %12s"
    fmt = "%12s  %12d  %12.1f  %12.3f  %12d"

    print("-" * (12*5 + 4*2))
    print(rowfmt % keys)
    print("-" * (12*5 + 4*2))
    for row in x:
        print(fmt % row)
    print("-" * (12*5 + 4*2))


def _package_data_file(filepath: str)-> str:
    try:
        import importlib.resources
        # importlib.resrouces.files is new in python 3.9
        # this is a workaround I found
        with importlib.resources.path("wordleaisql", filepath) as f:
            return str(f)
    except:
        import importlib_resources
        return str(importlib_resources.files("wordleaisql") / filepath)
    raise RuntimeError("File '{}' not found".format(filepath))

def _read_vocabfile(filepath: str)-> dict:
    assert os.path.isfile(filepath), "'{}' does not exist".format(filepath)
    opener = gzip.open if filepath.endswith(".gz") else open
    with opener(filepath, "rt") as f:
        out = {}
        for line in f:
            tmp = line.strip()
            if len(tmp) == 0:
                continue
            tmp = tmp.split()
            if len(tmp) == 1:
                # add weight 1
                out[tmp[0]] = 1
            else:
                v = float(tmp[1])
                if v < 0:
                    logger.warning("Negative weight is not allowed, '%s' is changed to zero", v)
                    v = 0
                out[tmp[0]] = v
    # some weight must be positive
    flg = any(p > 0 for p in out.values())
    #print(out.values())
    if not flg:
        raise ValueError("All weights are zero")
    return out

def default_vocabfile(level: int=3)-> str:
    return _package_data_file(os.path.join("vocab", "wordle-level{}.txt".format(level)))

def default_wordle_vocab(level: int=3)-> dict:
    vocabfile = default_vocabfile(level)
    words = _read_vocabfile(vocabfile)
    return words

@contextmanager
def _timereport(taskname: str="task", datetimefmt: str="%Y-%m-%d %H:%M:%S"):
    t1 = datetime.now()
    logger.info("Start %s (%s)", taskname, t1.strftime(datetimefmt))
    yield
    t2 = datetime.now()
    logger.info("End %s (%s, elapsed: %s)", taskname, t2.strftime(datetimefmt), t2-t1)


def _all_wordle_judges(words: list):
    total = len(words)**2
    nchar = len(str(total))
    for input_word, answer_word in tqdm(itertools.product(words, words), total=total):
        response = wordle_judge(input_word, answer_word)
        yield (input_word, answer_word, response)

def _compile_cpp(scriptfile: str, execfile: str, md5file: str, compiler: str=None, recompile: bool=False)-> bool:
    # Returns true is successful

    # we keep the md5 info of the source file to detect any changes
    # and compile the file only if the hash is not changed
    os.makedirs(os.path.dirname(execfile), exist_ok=True)
    os.makedirs(os.path.dirname(md5file), exist_ok=True)

    # compare the hash record
    if os.path.isfile(md5file):
        with open(md5file) as f:
            hash_prev = f.read()
    else:
        hash_prev = None
    h = hashlib.md5()
    with open(scriptfile, "rb") as f:
        h.update(f.read())
    hash_this = h.hexdigest()
    script_updated = (hash_this != hash_prev)
    
    if os.path.isfile(execfile) and (not script_updated) and (not recompile):
        logger.info("Compiled file ('%s') already exists and source has no update", execfile)
        return True
    else:
        # compile cpp script
        if compiler is None:
            # find a c++ compiler
            for c in ("g++", "clang++"):
                try:
                    subprocess.run([c, "--help"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    compiler = c
                    logger.info("C++ compiler detected: '%s'", compiler)
                    break
                except FileNotFoundError:
                    continue
        if compiler is None:
            logger.warning("No C++ compiler is found, so C++ enhancement is not available")
            return False

        logger.info("Compiling C++ script (%s)", scriptfile)
        try:
            subprocess.run([compiler, "-Wall", "-Werror", "-O3", "-o", execfile, scriptfile])
            with open(md5file, "w") as f:
                f.write(hash_this)
        except Exception as e:
            logger.warning("C++ compile failed, so C++ enhancement is not available")
            return False
    return True

def _prep_cpp(words: list, recompile: bool=False, compiler: str=None)-> str:
    # Returns:
    #   compiled executable path if cpp enhancement is available
    #   none otherwise

    # check if all words are ascii. otherwise the current C++ script is not applicable
    def _ascii_no_space(letter):
        if ord(letter) > 255 or ord(letter) < 1:
            return False
        return True
    for word in words:
        if not all(_ascii_no_space(letter) for letter in word):
            logger.info("Word contains non-ascii letter '%s'", word)
            return None

    scriptfile = _package_data_file("wordle-judge-all.cpp")
    #print(scriptfile)
    #execfile = os.path.abspath(os.path.join(os.path.dirname(__file__), "wordle-all-pairs.o"))
    execfile = os.path.expanduser("~/.worldaisql/wordle-judge-all.o")
    md5file = os.path.expanduser("~/.worldaisql/wordle-all-pairs.cpp.md5sum")

    if not _compile_cpp(scriptfile, execfile, md5file, compiler=compiler, recompile=recompile):
        logger.info("C++ compile failed")
        return None
    return execfile

def _all_wordle_judges_cpp(words: list, execfile: str):
    with TemporaryDirectory() as tmpdir:
        # create input file for the c++ script
        infile = os.path.join(tmpdir, "infile.txt")
        with open(infile, "w") as f:
            f.write(str(len(words)))
            f.write(" ".join(words))

        # run c++ script to save the results as a csv file
        outfile = os.path.join(tmpdir, "outfile.txt")
        #outfile = "responses.txt"  # for temporary check for the output table
        with open(infile) as f, open(outfile, "w") as g:
            with _timereport("Computing all wordle results"):
                subprocess.run([execfile], stdin=f, stdout=g)
        
        # generate the outcomes
        with open(outfile) as f:
            total = len(words)**2
            for line in tqdm(f, total=total):
                yield line.strip().split(" ")
    
def all_wordle_judges(words: list, use_cpp: bool=True, recompile: bool=False, compiler: str=None):
    if use_cpp:
        execfile = _prep_cpp(words, recompile, compiler)
        if execfile is not None:
            return _all_wordle_judges_cpp(words, execfile)
        else:
            logger.warning("C++ enhancement is not available, pure python implementation is used instead")

    return _all_wordle_judges(words)


# -*- coding: utf-8 -*-

"""
SQLite backend with no precomputation.

A quick version where the judge results are not precomputed.

Only table created:
    {vocabname}_words_approx   : contains all words
"""

import os
import sys
import math
import random
import sqlite3
from contextlib import contextmanager
from logging import getLogger
logger = getLogger(__name__)

from .utils import WordEvaluation, wordle_judge, _read_vocabfile, _dedup
from .sqlite import WordleAISQLite

@contextmanager
def _connect(db: str or sqlite3.Connection)-> sqlite3.Connection:
    if type(db) == sqlite3.Connection:
        yield db
    elif type(db) == str:
        conn = sqlite3.connect(db)
        try:
            yield conn
        finally:
            conn.close()
    else:
        raise TypeError("`db` must be either str or sqlite3.Connection, but '{}'".format(type(db)))

def _setup(db: str or sqlite3.Connection, vocabname: str, words: list or dict):
    assert len(words) == len(set(words)), "input_words must be unique"
    wordlens = set(len(w) for w in words)
    assert len(wordlens) == 1, "word length must be equal, but '{}'".format(wordlens)
    #with sqlite3.connect(dbfile) as conn:
    with _connect(db) as conn:
        c = conn.cursor()
        c.execute('DROP TABLE IF EXISTS "{name}_words_approx"'.format(name=vocabname))
        c.execute('CREATE TABLE "{name}_words_approx" (word TEXT PRIMARY KEY, weight FLOAT)'.format(name=vocabname))
        params = (
            words.items() if isinstance(words, dict) else
            [(w, 1) for w in words] if isinstance(words, list) else
            None
        )
        if params is None:
            raise TypeError("Unsupported type of `words`, '{}'".format(type(words)))
        c.executemany('INSERT INTO "{name}_words_approx" VALUES (?,?)'.format(name=vocabname), params)
        c.execute('CREATE INDEX "{name}_words_approx_idx" ON "{name}_words_approx" (word)'.format(name=vocabname))
        conn.commit()

# def _ensure_word_weight_column(dbfile: str, vocabname: str):
#     """If weight column is missing in the words table, add it with a constant 1"""
#     with sqlite3.connect(dbfile) as conn:
#         c = conn.cursor()
#         # check the existing column
#         c.execute('SELECT * FROM "{name}_words_approx" LIMIT 1'.format(name=vocabname))
#         exist = any(col[0].lower() == "weight" for col in c.description)
#         if not exist:
#             c.execute('ALTER TABLE "{name}_words_approx" ADD weight FLOAT'.format(name=vocabname))
#             logger.info('Added column `"%s_words_approx".weight`', vocabname)
#             c.execute('UPDATE "{name}_words_approx" SET weight = 1.0'.format(name=vocabname))
#             logger.info('Filled `"%s_words_approx".weight` with ones', vocabname)
#         conn.commit()

def _evaluate(db: str or sqlite3.Connection, vocabname: str, top_k: int=20, criterion: str="mean_entropy", candidates: list=None,
              word_pair_limit: int=500000, candidate_samplesize: int=500)-> list:
    assert candidate_samplesize > 0
    assert word_pair_limit > candidate_samplesize
    allwords = _words(db, vocabname)  # get all words

    n_words = len(allwords)
    n_candidates = n_words if candidates is None else len(candidates)
    candidate_samplesize = min(candidate_samplesize, n_candidates)  # can only upto the population size
    # make filters to input and answer words to conduct approx, smaller optimization
    if n_words * n_candidates <= word_pair_limit:
        # within the size limit, no need for approximation
        logger.debug("No approximation needed (input words: %d, candidates: %d)", n_words, n_candidates)
        inputfilter = ""
        params1 = ()
        if candidates is None:
            answerfiler = ""
            params2 = ()
        else:
            answerfilter = "WHERE word IN ({})".format(",".join("?" * n_candidates))
            params2 = tuple(candidates)
        params = tuple(params1) + tuple(params2)
    elif n_words * candidate_samplesize <= word_pair_limit:
        # need approximation, and
        # we can reduce the problem size by sampling the answer words only
        n_candidates2 = int(word_pair_limit / n_words)  # candidate sample size
        logger.debug("Approximation with candidate sampling (input words: %d, candidates: %d -> %d)",
                     n_words, n_candidates, n_candidates2)
        answerfilter = "WHERE word IN ({})".format(",".join("?" * n_candidates2))
        params2 = random.sample(allwords if candidates is None else candidates, n_candidates2)
        inputfilter = ""
        params1 = ()
        params = tuple(params1) + tuple(params2)
    else:
        # need approximation, and need input words sampling
        n_words2 = int(word_pair_limit / candidate_samplesize)
        inputfilter = "WHERE word IN ({})".format(",".join("?" * n_words2))
        params1 = random.sample(allwords, n_words2)
        if candidate_samplesize == n_candidates:
            logger.debug("Approximation with input word sampling (input words: %d -> %d, candidates: %d)",
                         n_words, n_words2, candidate_samplesize)
            if candidates is None:
                answerfiler = ""
                params2 = ()
            else:
                answerfilter = "WHERE word IN ({})".format(",".join("?" * n_candidates))
                params2 = tuple(candidates)
        else:
            logger.debug("Approximation with input word and candidate sampling (input words: %d -> %d, candidates: %d -> %d)",
                         n_words, n_words2, n_candidates, candidate_samplesize)
            answerfilter = "WHERE word IN ({})".format(",".join("?" * candidate_samplesize))
            params2 = random.sample(allwords if candidates is None else candidates, candidate_samplesize)
        params = tuple(params1) + tuple(params2)

#    with sqlite3.connect(dbfile) as conn:
    with _connect(db) as conn:
        conn.create_function("log2", 1, math.log2)
        conn.create_function("WordleJudge", 2, wordle_judge)
        c = conn.cursor()

        q = """
        with judges AS (
          SELECT
            a.word AS input_word,
            b.word AS answer_word,
            WordleJudge(a.word, b.word) AS judge
          FROM
            (SELECT word FROM {vocabname}_words_approx {inputfilter}) AS a,
            (SELECT word FROM {vocabname}_words_approx {answerfilter}) AS b
        ),
        tmp AS (
          SELECT
            input_word,
            judge,
            count(*) AS n,
            log2(count(*)) AS entropy
          FROM
            "judges"
          GROUP BY
            input_word, judge
        )
        SELECT
          input_word,
          max(n) AS max_n,
          1.0 * sum(n*n) / sum(n) AS mean_n,
          sum(n*entropy) / sum(n) AS mean_entropy
        FROM
          tmp
        GROUP BY
          input_word    
        """.format(vocabname=vocabname, inputfilter=inputfilter, answerfilter=answerfilter)
        #print(q)
        #print(len(params1), len(params2))
        #print(s)
        if len(params)==0:
            c.execute(q)
        else:
            c.execute(q, params)
        candidate_set = None if candidates is None else set(candidates)
        out = {row[0]: row + (1 if candidate_set is None else int(row[0] in candidate_set),) for row in c}
    # we pad random evals is there are insufficient rows
    # for padded words, we assign the worst possible values for max_n, mean_n, mean_entropy
    defaults = (n_candidates, n_candidates, math.log2(n_candidates))
    for w in allwords:
        if len(out) >= top_k:
            break
        if w in out:
            continue
        out[w] = (w,) + defaults + (int(w in candidate_set),)
    out = [WordEvaluation(*row) for row in out.values()]
    out.sort(key=lambda row: (getattr(row, criterion), -row.is_candidate))
    out = out[:top_k]
    return out

def _vocabnames(db: str or sqlite3.Connection)-> list:
#    with sqlite3.connect(dbfile) as conn:
    with _connect(db) as conn:
        c = conn.cursor()
        c.execute("SELECT name FROM sqlite_master")
        tables = [row[0] for row in c]
        #print(tables)
        t = [t[:-13] for t in tables if t.endswith("_words_approx")]
    return t

def _words(db: str or sqlite3.Connection, vocabname: str)-> list:
#    with sqlite3.connect(dbfile) as conn:
    with _connect(db) as conn:
        c = conn.cursor()
        c.execute('SELECT * FROM "{name}_words_approx"'.format(name=vocabname))
        words = [row[0] for row in c]
    return words

def _choose_word_with_weight(db: str or sqlite3.Connection, vocabname: str)-> str:    
    with _connect(db) as conn:
        sqlite3.enable_callback_tracebacks(True)
        conn.create_function("log", 1, math.log)
        c = conn.cursor()
        # The query below uses the fact that
        # Prob{ u_i^(1/w_i) > u_j^(1/w_j) } = w_i / (w_i + w_j),
        #   where u_i, u_j ~ uniform(0, 1).
        #
        # References: 
        #   https://stackoverflow.com/questions/1398113/how-to-select-one-row-randomly-taking-into-account-a-weight
        #   http://utopia.duth.gr/~pefraimi/research/data/2007EncOfAlg.pdf
        # 
        # On SQLite, random() generate an integer from [-9223372036854775808, +9223372036854775808]
        # Remove the sign and take mod n to make them roughly uniform on [0, n-1]
        # Then plus 0.5 to avoid zero, i.e. uniform on [0.5, ..., n-0.5]
        # Dividing by n, uniform on [0.5/n, ... 1-0.5/n]
        # With n large enough, the retult is roughly uniform on (0, 1)
        q = """
        SELECT
          word,
          weight / log((abs(random()) % 100000 + 0.5) / 100000.0) AS priority,
          weight
        FROM
          "{name}_words_approx"
        WHERE
          weight > 0
        ORDER BY priority
        LIMIT 1
        """.format(name=vocabname)
        c.execute(q)
        ans = c.fetchall()
    #print(ans)
    return ans[0][0]

def _weight_defined(db: str or sqlite3.Connection, vocabname: str)-> bool:
    with _connect(db) as conn:
        c = conn.cursor()
        # check the existing column
        c.execute('SELECT * FROM "{name}_words_approx" LIMIT 1'.format(name=vocabname))
        for col in c.description:
            if col[0].lower() == "weight":
                return True
    return False

class WordleAIApprox(WordleAISQLite):
    """
    Wordle AI with SQLite backend without precomputation.

    By omitting precomputation, this class computes approximation results
    faster and with smaller storage.

    Vocab information is stored in {vocabname}_words

    Args:
        vocabname (str):
            Name of vocaburary
        words (str or list or dict): 
            If str, the path to a vocabulary file
            If list, the list of words
            If dict, mapping from word to the weight
            Can be omitted if the vocabname is already in the database and resetup=False
        dbfile (str):
            SQLite database file
            If not supplied, use environment variable `WORDLEAISQL_DBFILE` if exists,
            otherwise './wordleai.db' in the current directory is used
        inmemory (bool):
            If true, ignore `dbfile` and use ':memory:' instead, so the in-memory database is used.

        word_pair_limit (int):
            Limit of the len(input_words) * len(candidates) to compute the evaluation.
            The larger, the more accurate approximation.
        candidate_samplesize (int):
            Sample size for the answer words for approximation.

        decision_metric (str):
            The criteria to pick a word
            Either 'max_n', 'mean_n', of 'mean_entropy'
        candidate_weight (float):
            The weight added to the answer candidate word when picking a word
        strength (float):
            AI strength in [0, 10]

        resetup (bool):
            Setup again if the vocabname already exists
    """
    def __init__(self, vocabname: str, words: list or str=None, dbfile: str=None, inmemory: bool=False,
                 word_pair_limit: int=500000, candidate_samplesize: int=500,
                 decision_metric: str="mean_entropy", candidate_weight: float=0.3, strength: float=6,
                 resetup: bool=False, **kwargs):
        if inmemory:
            dbfile = ":memory:"  # ignore dbfile supplied and use in-memory database
        else:
            if dbfile is None:
                dbfile = os.environ.get("WORDLEAISQL_DBFILE")
                if dbfile is None:
                    dbfile = "./wordleai.db"
            os.makedirs(os.path.dirname(os.path.abspath(dbfile)), exist_ok=True)
        self.dbfile = dbfile
        self.db = sqlite3.connect(self.dbfile)
        assert word_pair_limit > candidate_samplesize
        assert candidate_samplesize > 0
        self.word_pair_limit = word_pair_limit
        self.candidate_samplesize = candidate_samplesize
        self.vocabname = vocabname
        self.decision_metric = decision_metric
        self.candidate_weight = candidate_weight
        self.strength = min(max(strength, 0), 10)  # clip to [0, 10]
        # strength is linearly converted to the power of noise: 0 -> +5, 10 -> -5
        # larger noise, close to random decision
        self.decision_noise = math.pow(10, 5-self.strength)

        #print("vocabnames", self.vocabnames)
        if resetup or (vocabname not in self.vocabnames):
            assert words is not None, "`words` must be supplied to setup the vocab '{}'".format(vocabname)
            _words = (
                words if isinstance(words, dict) else
                {w:1.0 for w in words} if isinstance(words, list) else
                _read_vocabfile(words) if isinstance(words, str) else
                None
            )
            if _words is None:
                raise TypeError("Unsupported type 'words': '{}'".format(type(words)))
            logger.info("Setup tables for vocabname '%s'", vocabname)
            _setup(db=self.db, vocabname=vocabname, words=words)
        # else:
        #     _ensure_word_weight_column(dbfile, vocabname)  # make sure previously created words table has the weight column

        self._info = []                  # infomation of the judge results
        self._nonanswer_words = set([])  # words that cannot become an answer
        #self.set_candidates()

    def __del__(self):
        try:
            self.db.close()
            logger.debug("Database connection closed")
        except Exception as e:
            logger.warning("Failed to close the database connection: '%s'".format(e))

    @property
    def name(self)-> str:
        return "Wordle AI (SQLite backend, approx)"

    @property
    def vocabnames(self)-> list:
        """Available vocab names"""
        #return _vocabnames(self.dbfile)
        return _vocabnames(self.db)
    
    @property
    def words(self)-> list:
        """All words that can be inputted"""
        #return _words(self.dbfile, self.vocabname)
        return _words(self.db, self.vocabname)

    def evaluate(self, top_k: int=20, criterion: str="mean_entropy")-> list:
        """
        Evaluate input words and return the top ones in accordance with the given criterion
        """
        # return _evaluate(self.dbfile, self.vocabname, top_k=top_k, criterion=criterion, candidates=self.candidates,
        #                  word_pair_limit=self.word_pair_limit, candidate_samplesize=self.candidate_samplesize)
        return _evaluate(self.db, self.vocabname, top_k=top_k, criterion=criterion, candidates=self.candidates,
                         word_pair_limit=self.word_pair_limit, candidate_samplesize=self.candidate_samplesize)

    def choose_answer_word(self)-> str:
        """Randomly choose an answer word in accordance with the given weight"""
        if not _weight_defined(self.db, self.vocabname):
            print("Word weight is not defined. Please call `WordleAIApprox` with `resetup=True` next time", file=sys.stderr)
            return random.choice(self.words)
        return _choose_word_with_weight(self.db, self.vocabname)